import os, string, re, time, sys
import urllib, tempfile

# CountInstances
import subprocess
from subprocess import PIPE

# WaitForExit
import select, thread

# WineDoors global instances
from preferences import preferences
from log import log
from const import *
from utils import GetCDMountPoint
import commands
import urllib

class Wine:
    def __init__( self ):
        self.gui = True

    def WineVersion( self ):
        """ Returns wine version """
        path = os.popen(preferences['winebinary'] + " --version" ).readlines()
        for line in path:
            if line.find("pulseaudio") == -1:
                return line.rstrip().split("-")[1]

    def CountInstances( self, process_name="wine" ):
        process_name = process_name + " "
        log.Debug( "Counting instances of " + process_name )
        ps = subprocess.Popen( ["ps", "xwww"], stdout=PIPE )
        gp = subprocess.Popen( ["grep", process_name], stdin=ps.stdout, stdout=PIPE )
        gg = subprocess.Popen( ["grep", "-v", "grep"], stdin=gp.stdout, stdout=PIPE )
        gs = subprocess.Popen( ["grep", "-v", "python"], stdin=gg.stdout, stdout=PIPE )
        wc = subprocess.Popen( ["wc", "-l"], stdin=gs.stdout, stdout=PIPE )
        processes = wc.communicate()[0]
        processes = processes.rstrip()
        log.Debug( "Found "+ processes+" " + process_name + "processes active" )
        processes = int( processes )

        return processes
    
    def CheckDrive( self, drivepath=None ):
        self.drive = False
        if not drivepath:
            drivepath = preferences['wineroot']
        log.Log( "Checking wine drive: " + drivepath )
        if not os.path.isdir( drivepath ):
            log.Debug( "wine.py: CheckDrive: Didn't find folder " + drivepath )
        else:
            log.Debug( "wine.py: CheckDrive: Found folder " + drivepath )
        
        if not os.path.isdir( os.path.expanduser( "~/.wine-doors/" ) ):
            log.Debug( "wine.py: CheckDrive: Didn't find folder ~/.wine-doors/" )
        else:
            log.Debug( "wine.py: CheckDrive: Found folder ~/.wine-doors/" )
        
        confpath = os.path.join( preferences['wineroot'], "wine-doors/configured" )
        
        if not os.path.isfile( confpath ):
            log.Debug( "wine.py: CheckDrive: Didn't find file " + confpath)
        else:
            self.drive = True
            log.Debug( "wine.py: CheckDrive: Found file " + confpath)

        if not self.drive:
            log.Log( "wine.py: CheckDrive: No wine-drive defined in specified wineroot" )
            
        return self.drive
        
    def CreateDrive( self ):
        """ 
        Create a bare bones wine drive, we apply registry hacks and create
        the necessary files for this drive to behave like a wine-doors bottle.
        """
        if os.path.isdir( preferences['wineroot'] + "drive_c/" ):
            log.Log( "wine.py: CreateDrive: Wine drive already exists." )
            log.Debug( "         "+preferences['wineroot'] )
        else:
            log.Debug( "wine.py: CreateDrive: Wine Prefix Create "+preferences['wineroot'] )
            os.system( "wineprefixcreate --prefix %s -w > /dev/null 2>&1" % preferences['wineroot'] )
       
        if not os.path.isdir( preferences['wineroot'] ):
            log.Error( "wine.py: CreateDrive: Failed to create wine drive" )
            return
        
        log.Debug( "wine.py: CreateDrive: Creating wine-doors paths" )
        
        if not os.path.exists( os.path.expanduser( "~/.wine-doors" ) ):
            os.mkdir( os.path.expanduser( "~/.wine-doors" ) )
        if not os.path.exists( os.path.expanduser( "~/.wine-doors/installer-cache" ) ):
            os.mkdir( os.path.expanduser( "~/.wine-doors/installer-cache" ) )
        if not os.path.exists( os.path.expanduser( "~/.wine-doors/packlists" ) ):
            os.mkdir( os.path.expanduser( "~/.wine-doors/packlists" ) )
        if not os.path.exists( os.path.expanduser( "~/.wine-doors/apppacks" ) ):
            os.mkdir( os.path.expanduser( "~/.wine-doors/apppacks" ) )
        if not os.path.exists( os.path.expanduser( "~/.wine-doors/icons" ) ):
            os.mkdir( os.path.expanduser( "~/.wine-doors/icons" ) )
        # This is a hack for the home drive path, sometimes Y is expected
        if not os.path.exists(preferences['wineroot']+"dosdevices/y:"):
            os.system("ln -s "+os.path.expanduser("~")+" \""+preferences['wineroot']+"/dosdevices/y:\"")
        log.Debug( "wine.py: CreateDrive: Configuring drives" )
        os.system( "winecfg -D" ) # autoconfigure drives
        self.WaitForExit()

        log.Debug( "wine.py: CreateDrive: Registry changes" )

        self.setRegistry("HKLM\\Software\\Microsoft\\Windows\\CurrentVersion\\RegisteredOwner", preferences['name'] )
        self.setRegistry("HKLM\\Software\\Microsoft\\Windows\\CurrentVersion\\RegisteredOrganization", preferences['company'] )
        
        # Add any registry hacks that may be needed for normal operation
        self.Execute("regedit", preferences['prefix'] + "registry/audio-alsa.reg")
        self.WaitForExit()
        log.Log("wine.py: CreateDrive: Finished")
        if not os.path.isdir(preferences['wineroot']+"/drive_c/Program Files"):
            os.system("ln -s \""+self.ConvertPath(self.getProgramFilesDir())+"\" \""+
                      preferences['wineroot']+"/drive_c/Program Files\"")   
        os.system("touch "+ os.path.expanduser(os.path.join(preferences['wineroot'], "wine-doors/configured") ) )
        
    def WaitForExit( self, timeout=300 ):
        reader, writer = os.pipe()
        def wait_for_wineserver(timeout):
            done = False
            log.Debug( "Waiting for wineservers to exit" )
            for x in range( 1, timeout+1 ):
                if self.CountInstances( "wineserver" ) == 0:
                    done = True
                    break
                else:
                    time.sleep( 1 )
            if not done and self.CountInstances( "wineserver" ) > 0:
                log.Error( "Wineservers still running after %d seconds" % timeout )
            else:
                os.write( writer, "done" )
        thread.start_new_thread( wait_for_wineserver, ( timeout, ) )
        readable, dummy, dummy = select.select( ( reader, ), (), (), timeout )
        if readable:
            return True
        else:
            return False
    
    def Kill( self ):
        log.Debug( "Killing wine" )
        for pname in ( "wine", "wine-pthread", "wineserver", "winedbg" ):
            while self.CountInstances( pname ):
                log.Debug( "Killing %s" % pname )
                os.system( "/usr/bin/killall -9 %s" % pname )
        self.Reboot()
    
    def Reboot( self ):
        log.Debug( "Rebooting wine" )
        os.system( "%s wineboot -r" % preferences['winebinary'] )
        
    def isShellScript( self, script ):
        if script.endswith( ".sh" ):
            return True
        else:
            return False
            
        filetype = os.popen( "head -n 1 "+script ).read()
        if re.search ("^#!/", filetype):
            return True
        return False
        
    def ConvertPath( self, path ):
        """
        Convert a path from wine notation to linux absolute path
        """
        path = os.popen("winepath \""+ path +"\"" ).readlines()
        for line in path:
            if line.find("pulseaudio") == -1:
                return line.rstrip()
    
    def Execute( self, executable, argstring=None ):
        """
        Execute the installer, this will run regedit, msiexec, sh or wine
        depending on the type of file that's to be executed
        """
        ahk = False
        run = None
        oldcwd = False
        filefound = False

        if executable == "regedit" and not argstring:
            return

        if not argstring:
            argstring = ""
        else:
            argstring = " "+argstring 

        # fix file:// URI's
        executable = urllib.unquote( executable ).replace( "file://", "", 1 )
        exec_orig = executable

        # Avoid locking the CD drive, when ?:\\ is passed
        if re.search("^.:\\\\", executable):
            executable = self.ConvertPath( executable )
            oldcwd = os.getcwd()
            os.chdir( os.path.expanduser( "~/" ) )
        
        # Various tests to locate the file, if we still fail return with an error
        if not os.path.isfile(executable) and executable != "regedit" and executable != "winver":
            log.Warn("Can't find executable %s, Assuming CD drive path" % executable)
            cd = GetCDMountPoint()
            exec_orig = re.sub("\\\\\\\\", "/", exec_orig)
            exec_orig = re.sub("\\\\", "/", exec_orig)
            for thiscd in cd:
                log.Warn(thiscd)
                executable = re.sub("^.*:", str(thiscd), exec_orig)
                if os.path.isfile(executable):
                    filefound = True
                    break
            if not filefound:
                log.Warn("Executable file not found: " + executable)
                return FILE_NOT_FOUND

        # Determine the executable, and construct an executable command                
        if executable == "regedit":
            executable = "C:\\windows\\regedit.exe"
        elif executable == "winver":
            executable = "C:\\windows\\system32\\winver.exe"
        elif re.search("\.msi", executable):
            run = "msiexec /i " + executable
        elif re.search(".ahk", executable):
            run = preferences['winebinary'] + " \"" + \
                  self.getProgramFilesDir() + \
                  "\\AutoHotKey\\AutoHotKey.exe\" " + executable
            ahk = True
        elif self.isShellScript( executable ):
            run = "/bin/sh " + executable
            ahk = True

        if not run:
            run = "%s \"%s\"" % (preferences['winebinary'], executable)

        # Set useful env vars to pass to scripts etc...
        run = "PFWIN=\"" + self.getProgramFilesDir() + "\" " + \
              "PFUNIX=\"" + self.ConvertPath(self.getProgramFilesDir()) + "\" " + \
              "WDPREFIX=" + preferences['prefix'] + " " + \
              "WINEPREFIX=" + preferences['wineroot'] + " " + \
              run + argstring + \
              " >> ~/.wine.stdout.log 2>>~/.wine.stderr.log"

        if ahk:
            log.Debug ( "Setting the keyboard layout to \"us\" for AutoHotKey" )
            os.system( "setxkbmap us" )

        log.Debug( "Executing: " + run )
        retval = os.system( run )

        if ahk:
            log.Debug ( "Resetting the default keyboard layout" )
            os.system( "/bin/bash " + os.path.dirname(os.path.realpath( __file__ )) + "/setdefaultxkbmap.sh" )

        if oldcwd:
            os.chdir( oldcwd )
        return retval

    def getRegistry( self, path, key=None ):
        """
        Expects a path in the form of HKLM\\\\..\\\\.. where HKLM = System.reg
        and HKCU = User.reg, all in all this is a pretty nasty hack :/
        TODO: Input should be sanitised someone, expect the unexpected!
        """
        if path.find("HKCU") != -1:
            if os.path.isfile( preferences['wineroot']+"/user.reg" ):
                reg_file = preferences['wineroot']+"/user.reg"
                path = path.replace("HKCU\\\\", "")
            else:
                log.Error("Cannot locate user.reg")
                return
        
        if path.find("HKLM") != -1:
            if os.path.isfile( preferences['wineroot']+"/system.reg" ):
                reg_file = preferences['wineroot']+"/system.reg"
                path = path.replace("HKLM\\\\", "")
            else:
                log.Error("Cannot locate system.reg")
                return
        if not key:
            reg = re.compile("(.*)\\\\(.*)\Z")
            m = reg.match(path)
            path = m.group(1)
            key = m.group(2)
        data = ""
        inkey = False
        for line in file(reg_file):
            if line.find(path) != -1:
                inkey = True
            elif inkey:
                if line.find("[") != -1:
                    break
                if line.find("=") == -1:
                    continue
                line = line.replace( "\"", "")
                line = line.replace( "\n", "")
                [thiskey, thisval] = line.split( "=")
                if thiskey == key:
                    data = thisval
                    break
        return data.replace("dword:", "", 1)

    def setRegistry(self, path, value):
        """
        This is a hack for setting individual registry keys
        it should make it relatively easy to implement lots of 
        registry hacks in the UI...
        We clean up the input and will accept \\\\ or \\ and
        we strip any square brackets too, then we generate
        a regedit 4 file then pass it over to the execute method
        """
        path.replace("\\\\", "\\")
        path.replace("[", "")
        path.replace("]", "")
        if path.find("HKLM") != -1:
            path = path.replace("HKLM", "HKEY_LOCAL_MACHINE")
        if not path.find("HKCU") != -1:
            path = path.replace("HKCU", "HKEY_CURRENT_USER")

        reg = re.compile("(.*)\\\\(.*?)\Z")
        m = reg.match(path)
        path = m.group(1)
        key = m.group(2)

        __, tmp_name = tempfile.mkstemp(suffix='.reg') 
        f = open(tmp_name , "wt" )
        f.write("REGEDIT 4\n\n")
        f.write("["+path+"]\n")
        f.write("\""+key+"\"=\""+value+"\"\n")
        f.close()
        self.Execute("regedit", tmp_name);
        os.remove(tmp_name)

    def lsReg( self, path):
        """
        Expects a path in the form of HKLM\\\\..\\\\.. where HKLM = System.reg
        and HKCU = User.reg, all in all this is another pretty nasty hack :/
        """
        if path.find("HKCU") != -1:
            if os.path.isfile( preferences['wineroot']+"/user.reg" ):
                reg_file = preferences['wineroot']+"/user.reg"
                path = path.replace("HKCU\\\\", "")
            else:
                log.Error("Cannot locate user.reg")
                return
        
        if path.find("HKLM") != -1:
            if os.path.isfile( preferences['wineroot']+"/system.reg" ):
                reg_file = preferences['wineroot']+"/system.reg"
                path = path.replace("HKLM\\\\", "")
            else:
                log.Error("Cannot locate system.reg")
                return
        line_start = re.compile(".*\\\\")
        line_end = re.compile("].*\\n")
        data = []
        for line in file(reg_file):
            if line.find(path) != -1 and line_start.sub( "", line_end.sub( "", path) ) != line_start.sub( "", line_end.sub( "", line) ):
                data.append( line_start.sub( "", line_end.sub( "", line) ) )
        return data

    def getProgramFilesDir( self ):
        return self.getRegistry("HKLM\\\\Software\\\\Microsoft"+
                          "\\\\Windows\\\\CurrentVersion", 
                          "ProgramFilesDir")

    def installedFromReg( self , Args=None):
        """ 
        Returns a list of dicts containing info about installed apps from registry 
        """
        approot = "HKLM\\\\Software\\\\Microsoft\\\\Windows\\\\CurrentVersion\\\\Uninstall\\\\"
        apps = self.lsReg( approot )
        applist = []
        for app in apps:
            info = {}
            if self.getRegistry( approot + app, "FriendlyName" ) != "":
                info[u"name"] = self.getRegistry( approot + app, "FriendlyName" )
            elif self.getRegistry( approot + app, "DisplayName" ) != "":
                info[u"name"] = self.getRegistry( approot + app, "DisplayName" )
            else:
                continue
            # If description, use it.
            if self.getRegistry( approot + app, "Comments" ) != "":
                info[u"description"] = self.getRegistry( approot + app, "Comments" )
            else:
                info[u"description"] = "No description found"
            # Set icon
            if self.getRegistry( approot + app, "DisplayIcon" ) == "" or \
            self.getRegistry( approot + app, "DisplayIcon" ).lower().endswith( ".exe" ):
                info[u"icon"] = "http://www.wine-doors.org/wd-icons/ms-library.svg"
            else:
                info[u"icon"] = wine.ConvertPath( "'" + self.getRegistry( approot + app, "DisplayIcon") + "'")
            # Tidy up Uninstaller
            info[u"UninstallString"] = self.getRegistry( approot + app, "UninstallString" )
            if info[u"UninstallString"].startswith( "\\" ):
                info[u"UninstallString"] = info[u"UninstallString"].replace( "\\", "", 1 )
            if info[u"UninstallString"].startswith( "str(2):" ):
                info[u"UninstallString"] = info[u"UninstallString"].replace( "str(2):", "", 1 )
            rundll = re.compile("rundll32(\.exe|) ", re.IGNORECASE)
            info[u"UninstallString"] = rundll.sub("", info[u"UninstallString"])
            # Other info
            info[u"ms-license"] = "False"
            info[u"shortname"] = info[u"name"].lower().replace( " ", "_" )
            info[u"category"] = u"Applications > Non-WDI"
            info[u"default_version"] = self.getRegistry( approot + app, "DisplayVersion" )
            info[u"vendor"] = self.getRegistry( approot + app, "Publisher" )
            info[u"versions"] = [[(
                self.getRegistry( approot + app, "VersionMajor" ).replace("0", "") , #Major version FIXME
                u'0.9.0', # Wine version
                u'winehq', # Wine variant
                u'Gold', # Appdb rating
                u'None', # Installer size
                u'None', # Installed size
                u'5', # Votes
                self.getRegistry( approot + app, "DisplayVersion" ), # Version
                u"reg" # Pack
                )]] 
            info[u"requires"] = []
            info["repo"] = ["", ""]
            info[u"homepage"] = self.getRegistry( approot + app, "URLInfoAbout" )
            applist.append(info)
        return applist

    def installedInReg( self, pack ):
        self.installedFromReg()
        for installed in self.installedFromReg():
            if installed['shortname'] == pack:
                return True
        return False

    def getUninstallerFromReg( self, pack ):
        self.installedFromReg()
        for installed in self.installedFromReg():
            if installed[u'shortname'] == pack:
                return installed[u'UninstallString']
        return False
            
wine = Wine()

if __name__ == "__main__":
    wine.CheckDrive()
    if not wine.drive:
        print "Wine Drive does not exist, creating one"
        wine.CreateDrive()

    proc = wine.CountInstances()
    if proc > 0:
       print "Counted %d wine processes, waiting for exit" % proc
       wine.WaitForExit()
